##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import pandas as pd
from os.path import dirname
from os.path import join
import numpy as np

class ClassLabelException(Exception):
	pass

class DataSets():
	def __init__(self):
		pass

	def load_activity_raw(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'activity_raw.csv'))

	def load_activity_raw_toy(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'activity_raw_toy.csv'))

	def load_activity_norm(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'activity_norm.csv'))

	def load_activity_quant(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'activity_quant.csv'))

	def load_activity_quant_toy(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'activity_quant_toy.csv'))

	def load_gesture_raw(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'gesture_raw.csv'))

	def load_toy_data(self):
		module_path = dirname(__file__)
		return pd.read_csv(join(module_path, 'data', 'toy_data.csv'))

	@staticmethod
	def generate_step_data(window_size=100, num_classes=5, num_iterations=2, scale_factor=100, noise_scale=5, class_labels=None, class_values=None):
		"""Generate a labeled test data vector.

		Args:
		    window_size (int, optional): The size of each segment window.
		    num_classes (int, optional): The number of separate classes
		    num_iterations (int, optional): The number of iterations to create.
		    scale_factor (int, optional): scale the values of the data by this much before adding noise
		    noise_scale (int, optional): The range of noise to add (-noise_scale:noise_scale)
		    class_labels (None, optional): you can pass specific class labels as an array, otherwise defaults to
		     the range(num_classes)

		Returns:
		    DataFrame: a test data vector for building and testing models
		"""
		class_labels = check_class_label(num_classes, class_labels)
		class_values = check_class_values(num_classes, class_values)
		delta = window_size*num_classes
		signal_length = num_iterations*delta

		y = np.zeros(signal_length)
		y_label = np.zeros(signal_length)

		for j in range(num_iterations):
		    y_temp = np.zeros(window_size*num_classes)
		    if j%2==0:
		        for i in range(num_classes):

		            y_temp[window_size*i:window_size*(i+1)] = class_values[i]
		            y_label[window_size*i+j*delta:window_size*(i+1)+j*delta] = class_labels[i]
		    else:
		        for i in range(num_classes):
		            y_temp[window_size*i:window_size*(i+1)] = class_values[-1-i]
		            y_label[window_size*i+j*delta:window_size*(i+1)+j*delta] = class_labels[num_classes-1-i]
		    y[delta*j:delta*(j+1)] = y_temp*scale_factor

		if noise_scale > 0:
		      y += np.array([np.random.choice(range(-noise_scale,noise_scale)) for i in range(len(y))])

		return pd.DataFrame({'Data':y.astype(np.int16), 'Label':y_label.astype(int)})

	@staticmethod
	def generate_harmonic_data(window_size=100, num_classes=5, num_iterations=2, scale_factor=100, noise_scale=5, class_labels=None, harmonic_nodes=None):
		"""Generate a labeled test data sin vector.

		Args:
		    window_size (int, optional): The size of each segment window.
		    num_classes (int, optional): The number of separate classes
		    num_iterations (int, optional): The number of iterations to create.
		    scale_factor (int, optional): scale the values of the data by this much before adding noise
		    noise_scale (int, optional): The range of noise to add (-noise_scale:noise_scale)
		    class_labels (None, optional): you can pass specific class labels as an array, otherwise defaults to
		     the range(num_classes)

		Returns:
		    DataFrame: a test data vector for building and testing models
		"""

		class_labels = check_class_label(num_classes, class_labels)
		harmonic_nodes = check_class_values(num_classes, harmonic_nodes)

		delta = window_size*num_classes
		signal_length = num_iterations*delta

		y = np.zeros(signal_length)
		x = np.arange(signal_length)

		y_label = np.zeros(signal_length)
		for j in range(num_iterations):
		    y_temp = np.zeros(delta)
		    x = np.arange(window_size)
		    if j%2==0:
		        for i in range(num_classes):
		            y_temp[window_size*i:window_size*(i+1)] = get_standing_wave(x, window_size, harmonic_nodes[i])
		            y_label[window_size*i+j*delta:window_size*(i+1)+j*delta] = class_labels[i]
		    else:
		        for i in range(num_classes):
		            y_temp[window_size*i:window_size*(i+1)] = get_standing_wave(x, window_size, harmonic_nodes[-1-i])
		            y_label[window_size*i+j*delta:window_size*(i+1)+j*delta] = class_labels[num_classes-1-i]
		    y[delta*j:delta*(j+1)] = y_temp*scale_factor

		if noise_scale > 0:
		      y += np.array([np.random.choice(range(-noise_scale,noise_scale)) for i in range(len(y))])

		return pd.DataFrame({'Data':y.astype(np.int16), 'Label':y_label.astype(int)})


def get_standing_wave(x, L, n):
	ld = 2. * L / n
	return np.sin(x*(2*np.pi) / ld)

def check_class_label(num_classes, class_labels=None):
	if class_labels is None:
		class_labels = range(1, num_classes+1)
	elif len(class_labels) != num_classes:
		raise ClassLabelException("The number of classes is not equal to the number of class labels.")

	for i in class_labels:
		if type(i) != type(1):
			raise ClassLabelException("class labels must be integers.")
	if i <= 0:
			raise ClassLabelException("class labels must be greater than zero.")

	return class_labels

def check_class_values(num_classes, class_values=None):
	if class_values is None:
		class_values = range(1,num_classes+1)
	elif len(class_values) != num_classes:
		raise ClassLabelException("The number of classes is not equal to the number of class values.")

	for i in class_values:
		if type(i) != type(1):
			raise ClassLabelException("class values must be type int")
		if i <= 0:
			raise ClassLabelException("class values must be greater than zero.")

	return class_values


##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json

from six import string_types
import sensiml.base.utility as utility
from sensiml.datamanager.function import Function
from sensiml.method_calls import FeatureFileCall, QueryCall, FunctionCall, \
     TrainAndValidationCall, GeneratorCall, GeneratorCallSet, SelectorCall, \
     SelectorCallSet, ValidationMethodCall, ClassifierCall, OptimizerCall, \
     TrainingAlgorithmCall, CaptureFileCall

class FunctionExistsError(Exception):
    """Base class for a function exists error"""
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class Functions():
    """Base class for a collection of functions"""
    def __init__(self, connection):
        self._connection = connection
        self._function_list = {}
        self.build_function_list()

    @property
    def function_list(self):
        if not self._function_list:
            self.build_function_list()

        return self._function_list

    @function_list.setter
    def function_list(self, value):
        self._function_list = value


    def build_function_list(self):
        """Populates the function_list property from the server."""
        function_list = {}

        function_response = self.get_functions()
        for function in function_response:
            function_list[function.name] = function

        self.function_list = function_list

    def get_function_by_name(self, name):
        """Gets a function from the server by name.

            Args:
                name (str): function name

            Returns:
                function
        """
        return self.function_list.get(name, None)

    def _new_function_from_dict(self, dict):
        """Creates and populates a function from a set of properties.

            Args:
                dict (dict): contains properties of a function

            Returns:
                function
        """
        function = Function(self._connection)
        function.initialize_from_dict(dict)
        return function

    def get_functions(self, function_type=''):
        """Gets all functions as function objects.

            Args:
                function_type (optional[str]): type of function to retrieve

            Returns:
                list of functions
        """
        url = 'transform/'
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        functions = []
        for function_params in response_data:
            functions.append(self._new_function_from_dict(function_params))

        if function_type:
            functions = [f for f in functions if f.type == function_type]

        return functions

    def create_query_call(self, name='Query'):
        """Creates a query call.

            Returns:
                QueryCall
        """
        qy = QueryCall(name)
        return qy

    def create_featurefile_call(self, name):
        """Creates a featurefile call.

            Args:
                name: name of the featurefile

            Returns:
                FeatureFileCall
        """
        ff = FeatureFileCall(name)
        return ff

    def create_capturefile_call(self, name):
        """Creates a capturefile call.

            Args:
                name: name of the featurefile

            Returns:
                FeatureFileCall
        """
        ff = CaptureFileCall(name)
        return ff


    def create_train_and_validation_call(self, name='TVO'):
        """Creates an empty train and validation call.

            Returns:
                TrainAndValidationCall
        """
        tvo = TrainAndValidationCall()
        tvo.name = name
        return tvo

    def get_functions_by_type(self, function_type='', subtype=''):
        """Gets all functions or functions of a particular function ype or all functions
            of a particular subtype as function objects.

            Args:
                subtype (optional[str]): subtype to retrieve

            Returns:
                list of functions
        """
        # Populate each function from the server

        if function_type:
            functions = [f for f in self.function_list.values() if f.type == function_type]

        elif subtype:
            functions = [f for f in self.function_list.values() if f.subtype == subtype]

        else:
            functions = [f for f in self.function_list.values()]

        return functions

    def __str__(self, function_type=''):
        output_string = ''
        all_functions = self.get_functions_by_type(function_type=function_type)
        for s in set([f.type for f in all_functions]):
            output_string += '{0}:\n'.format(s)
            for t in [tr for tr in all_functions if tr.type == s and (tr.has_c_version or
                        tr.type not in ['Transform', 'Segmenter', 'Feature Generator'])]:
                if t.subtype:
                    output_string += '    {0} ({1})\n'.format(t.name, t.subtype)
                else:
                    output_string += '    {0}\n'.format(t.name)
            output_string += '\n'

        return output_string

    def __call__(self, function_type):
        return self.__str__(function_type)

    def __getitem__(self, key):
        return self.get_function_by_name(key).__str__()

    def _initialize_docstring(self, function):
        ic_dict = function.input_contract
        #oc_dict = function.output_contract
        inputs = {i['name']: None for i in ic_dict if not i.get('handle_by_set', False)} # Parse arguments from the input contract

        # Construct a user-friendly docstring using the description and input/output contracts
        docstring = function.description
        docstring += '\n \nInputs\n---------- \n'

        for i in ic_dict:
            if i.get('default', None):
                docstring += '  {0}: {1} - (default: {2})\n'.format(i['name'], i['type'], i['default'])
            else:
                docstring += '  {0}: {1} \n'.format(i['name'], i['type'])

        # docstring += '\nOutputs\n---------- \n'
        # for i in oc_dict:
        #     docstring += '  {0} \n'.format(i['type'])

        docstring += '\nUsage\n---------- \n'
        docstring += 'For DataFrame inputs, provide a string reference to the\nDataFrame output of a previous step in the pipeline.\n'
        docstring += 'For Dataframe output, provide a string name that subsequent\noperations can refer to.'

        return ic_dict, inputs, docstring

    def _warn_if_not_transcoded(self, function):
        if function.type in ('Transform', 'Segmenter', 'Feature Generator') and not function.has_c_version:
            print('Warning: {0} is not supported by KnowledgePack code generation. \n' \
            'It can be used for cloud-based modeling but cannot be downloaded or flashed to a device. \n'.format(function.name))

    def create_function_call(self, function, name = None):
        """Creates a function call object.

            Args:
                function (Function object or str): the function for which to create a call object or its
                string name

            Returns:
                function call object or None if the function is not found

            Raises:
                error if no call object can be initialized for the function type
        """
        if name:
            print("The use of the second input parameter is no longer necessary and will be deprecated.")

        if isinstance(function, string_types):
            function = self.function_list.get(function, None)

            if  function is None:
                print('No function found')
                return None

        self._warn_if_not_transcoded(function)

        try:
            return getattr(self, 'create_{}_call'.format(function.type.lower().replace(' ','_')))(function)
        except Exception as e:
            print('No function call for function type "{}" found.'.format(function.type.lower()))
            raise(e)

    def _create_base_call(self, function, CallFunction, ignore_columns=[]):
        """Creates a base subclass on the fly for the specified function.

        Parses the passed function's I/O contracts and builds class properties accordingly. Returns an instance of the
        new class.
        """

        if isinstance(function, string_types):
            function = self.function_list[function]

        self._warn_if_not_transcoded(function)
        ic_dict, inputs, docstring = self._initialize_docstring(function)

        for column in ignore_columns:
            inputs.pop(column, None)

        for contract in ic_dict:
            if 'default' in contract.keys():
                inputs[contract['name']] = contract['default']

        # Add docstring to inputs structure and create the new class
        inputs['__doc__'] = docstring
        fname = str(function.name)

        Call = type(fname, (CallFunction,), inputs)

        return Call(function.name, function_type=function.type.lower())

    def create_transform_call(self, function, name=None):
        """Creates a FunctionCall for the specified transform.

        Parses the passed function's I/O contracts and builds class properties accordingly. Returns an instance of the
        new class.

            Args:
                function (Function): function with which to create a transform call

            Returns:
                instance of FunctionCall customized for the input function
        """

        ic_dict, inputs, docstring = self._initialize_docstring(function)

        # Add docstring to inputs structure and create the new class
        inputs['__doc__'] = docstring
        fname = str('kb_'+function.name.lower().replace(' ', '_'))

        for contract in ic_dict:
            if 'default' in contract.keys():
                inputs[contract['name']] = contract['default']

        Call = type(fname, (FunctionCall,), dict({'_input_contract': ic_dict}, **inputs))
        return Call(name=function.name, function_type=function.type.lower())

    def create_segmenter_call(self, function, name=None):

        return self.create_transform_call(function, name=name)

    def create_sampler_call(self, function, name=None):

        return self.create_transform_call(function, name=name)

    def create_feature_generator_call(self, function):
        """Creates a GeneratorCall for the specified generator.

            Args:
                function (Function): function with which to create a feature generator call

            Returns:
                instance of GeneratorCall customized for the input function
        """
        ignore_columns = ['group_columns', 'input_data']

        return self._create_base_call(function, GeneratorCall, ignore_columns=ignore_columns)

    def create_feature_selector_call(self, function):
        """Creates a SelectorCall for the specified selector.

            Args:
                function (Function): function with which to create a feature selector call

            Returns:
                instance of SelectorCall customized for the input function
        """
        ignore_columns = ['ignore_columns', 'input_data', 'label_column', 'num_features']

        return self._create_base_call(function, SelectorCall, ignore_columns=ignore_columns)

    def create_validation_method_call(self, function, name = None):
        """Creates a ValidationMethodCall for the specified function.

            Args:
                function (Function): function with which to create a validation method call

            Returns:
                instance of ValidationMethodCall customized for the input function
        """
        return self._create_base_call(function, ValidationMethodCall)


    def create_classifier_call(self, function):
        """Creates a ClassifierCall for the specified function.

            Args:
                function (Function): function with which to create a classifier call

            Returns:
                instance of ClassifierCall customized for the input function
        """
        return self._create_base_call(function, ClassifierCall)


    def create_optimizer_call(self, function, name = None):
        """Creates an OptimizerCall for the specified function.

            Args:
                function (Function): function with which to create an optimizer call

            Returns:
                instance of OptimizerCall customized for the input function
        """
        return self._create_base_call(function, OptimizerCall)

    def create_generator_subtype_call(self, subtype):
        """Creates a GeneratorCall for the specified subtype of generators.

        Uses the input contract of one function of the subtype (assumed to be constant for all generators of the subtype),
        and builds class properties accordingly.

            Args:
                subtype (str): name of the subtype for which to create a generator call

            Returns:
                instance of GeneratorCall customized for the subtype
        """

        function_list = self.get_functions_by_type(subtype=subtype)
        if len(function_list) > 0:
            function = function_list[0]
            ic_dict, inputs, docstring = self._initialize_docstring(function)
        else:
            print( "Error: No generators of subtype {0}; length of function list {1}".format(subtype, len(function_list)))
            return None

        for function in function_list:
            self._warn_if_not_transcoded(function)

        ignore_columns = ['group_columns', 'input_data']

        for column in ignore_columns:
            inputs.pop(column, None)

        for contract in ic_dict:
            if 'default' in contract.keys():
                inputs[contract['name']] = contract['default']

        # Add docstring to inputs structure and create the new class
        inputs['__doc__'] = docstring
        inputs['_subtype'] = subtype

        Call = type(subtype, (GeneratorCall,), inputs)
        return Call(function.name, 'generator')

    def create_generator_call_set(self, name='GeneratorCallSet'):
        """Creates an empty generator call set."""
        gcs = GeneratorCallSet(name)
        return gcs

    def create_selector_call_set(self, name='SelectorCallSet'):
        """Creates an empty selector call set."""
        scs = SelectorCallSet(name)
        return scs

    def create_training_algorithm_call(self, function, name = None):
        """Creates a training algorithm method call for the specified function.

            Args:
                function (Function): function with which to create an training algorithm call

            Returns:
                instance of TrainingAlgorithmCall customized for the input function
        """
        return self._create_base_call(function, TrainingAlgorithmCall)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################
import pandas as pd
from pandas import DataFrame
import os.path

EXCEPTION_STATES = frozenset(['FAILURE', 'RETRY', 'REVOKED'])

# Custom ready states because sync or nonexistent task state is set to NONE
READY_STATES = frozenset(['FAILURE', 'REVOKED', 'SUCCESS', None])

# Custom unready states because celery's PENDING state is unreliable
UNREADY_STATES = frozenset(['STARTED', 'RECEIVED', 'RETRY', 'SENT'])

from sensiml.datamanager.metadataset import MetadataSet
from sensiml.datamanager.errors import CaptureUploadFailureError
import sensiml.base.utility as utility
from requests.exceptions import *
import time
import logging

logger = logging.getLogger(__name__)

class CaptureTaskFailedError(Exception):
    """Raised if Capture task failed"""

class Capture(object):
    """Base class for an Capture."""

    def __init__(self, connection, project, filename='', path='', uuid='', capture_info=None, **kwargs):
        """Initialize an Capture instance."""
        self._connection = connection
        self._project = project
        self._filename = filename
        self._path = path
        self._uuid = uuid
        self._capture_info = capture_info
        self._metadata = MetadataSet(self._connection, self._project, self)

    @property
    def uuid(self):
        """Auto generated unique identifier for the Capture object"""
        return self._uuid

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value

    @property
    def path(self):
        """The local or server path to the Capture file data"""
        return self._path

    @path.setter
    def path(self, value):
        self._path = value
    @property
    def metadataset(self):
        return self._metadata

    @property
    def capture_info(self):
        """Info about capture"""
        return self._capture_info

    @capture_info.setter
    def capture_info(self, value):
        self._capture_info = {}

        if value.get('CalculatedSampleRate', None) is not None:
            self._capture_info['calculated_sample_rate']  = value.get('CalculatedSampleRate')
        if value.get('SampleRate', None) is not None:
            self._capture_info['set_sample_rate'] = value.get('SampleRate',None)

        if not self._capture_info:
            self._caputre_info = None

    @property
    def ready(self):
        """Returns if Capture (Capture) is ready or not

        Returns:
            Boolean: True if task is ready or False if task is pending. Raises Exception if task failed.
        """
        url = 'project/{0}/capture/{1}/'.format(self._project.uuid, self.uuid)
        response = self._connection.request('get', url)
        response_data, err = utility.check_server_response(response)
        if not response_data['task_state'] or response_data['task_state'] in READY_STATES:
            return True

        if response_data['task_state'] in EXCEPTION_STATES:
            logger.error(response_data['task_result'])
            raise CaptureTaskFailedError()

        return False

    def await_ready(self, sleep=3, retries=0):
        """Blocks until Capture (Capture) is ready or failed

        Args:
            sleep (int): Number of seconds to sleep between polling
            retries (int): Number of times to retry before unblocking and returning False.
                           Use 0 for infinite.

        Returns:
            None or raises CaptureUploadFailureError if upload failed.
        """
        try_ = 0
        url = 'project/{0}/capture/{1}/'.format(self._project.uuid, self.uuid)

        while retries is 0 or try_ <= retries:
            try_ += 1
            response = self._connection.request('get', url)
            response_data, err = utility.check_server_response(response)
            # Uses custom UNREADY_STATES because we do NOT want to sleep forever on "PENDING"
            if response_data['task_state'] in UNREADY_STATES:
                time.sleep(sleep)
            elif not response_data['task_state'] or response_data['task_state'] in READY_STATES:
                return True
            else:
                raise CaptureUploadFailureError(response_data['task_result'])

    def insert(self, asynchronous=False):
        """Calls the REST API to insert a new Capture."""
        url = 'project/{0}/capture/'.format(self._project.uuid)
        capture_info = {'name': self.filename, 'async': asynchronous}
        if self.capture_info is not None:
            capture_info.update(self._capture_info)
        response = self._connection.file_request(url, self.path, capture_info, "rb")
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._uuid = response_data['uuid']

    def update(self):
        """Calls the REST API to update the capture."""
        url = 'project/{0}/capture/{1}/'.format(self._project.uuid, self._uuid)

        capture_info = {'name': self.filename}
        if self._capture_info is not None:
            capture_info.update(self._capture_info)
        if self.path:
            response = self._connection.file_request(url, self.path, capture_info, 'rb', 'patch')
        else:
            response = self._connection.request(url, 'post', capture_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API to delete the capture from the server."""
        url = 'project/{0}/capture/{1}/'.format(self._project.uuid, self._uuid)
        response = self._connection.request('delete', url,)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and self populates properties from the server."""
        url = 'project/{0}/capture/{1}/'.format(self._project.uuid, self._uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response( response )
        if err is False:
            self.filename = response_data['name']

    @classmethod
    def initialize_from_dict(cls, capture_dict):
        """Reads a dictionary or properties and populates a single capture.

            Args:
                capture_dict (dict): contains the capture's 'name' property

            Returns:
                capture object
        """
        assert isinstance(capture_dict, dict)
        new_dict = capture_dict.copy()
        new_dict['filename'] = str(new_dict.get(
            'filename',  # capture object attribute
            new_dict.pop('name', '')  # API object key
        ))
        return Capture(**new_dict)

    def data(self):
        """Calls the REST API and retrieves an capture's sensor data.

            Returns:
                data (DataFrame): all sensor data associated with the capture
        """
        url = 'project/{0}/capture/{1}/capture-sample/'.format(
            self._project.uuid, self.uuid)
        data = []
        try:
            paged_response = self._connection.paged_request('get', url)
            for results in paged_response:
                data.append(DataFrame(results))
        except HTTPError:
            utility.read_response_json_data(paged_response.response)
            raise
        data = pd.concat(data)


        # Post-processing to get numeric index
        data['sequence'] = data.sequence.astype(int)
        data = data.sort_values(by='sequence').reset_index(drop=True)
        data.drop('sequence', axis=1, inplace=True)

        return data

    ## methods to make this instance hashable and comparable
    def __hash__(self):
        return hash(self._uuid)
    def __eq__(self, other):
        return ((self._uuid,) == (other._uuid,))
    def __ne__(self, other):
        return not self == other
    def __lt__(self, other):
        return ((self._uuid,) < (other._uuid,))
    def __gt__(self, other):
        return ((self._uuid,) > (other._uuid,))
    def __le__(self, other):
        return (self < other) or (self == other)
    def __ge__(self, other):
        return (self > other) or (self == other)




from sensiml.datamanager.functions import Functions
from sensiml.datamanager.platformdescriptions import PlatformDescriptions
from sensiml.datamanager.pipelineseeds import  PipelineSeeds
from sensiml.datamanager.featurefiles import FeatureFiles
from sensiml.datamanager.projects import Projects
from sensiml.datamanager.segmenter import Segmenters



__all__ = ['Functions', 'FeatureFiles', 'Projects', 'Segmenters', 'PipelineSeeds','PlatformDescriptions']


class DataManagerError(Exception):
    pass

class AsyncObjectError(DataManagerError):
    pass

class AsyncObjectNotReadyError(AsyncObjectError):
    pass

class CaptureNotReadyError(AsyncObjectNotReadyError):
    pass

class AsyncObjectFailureError(AsyncObjectError):
    """When object processing/upload fails"""
    pass

class CaptureUploadFailureError(AsyncObjectFailureError):
    pass

# Renamed to event

class EventNotReadyError(CaptureNotReadyError):
    """Alias for CaptureNotReadyError since Captures
        are called Events in some locations."""
    pass

class EventUploadFailureError(CaptureUploadFailureError):
    pass